#!/usr/bin/env python3

import json
import random
import os
import signal
import sys
import time
import argparse

from dwq import Job, Disque
import dwq.util as util

from dwq.version import __version__

def sigterm_handler(signal, stack_frame):
    raise SystemExit()

def nicetime(time):
    secs = round(time)
    minutes = secs/60
    hrs = minutes/60
    days = int(hrs/24)
    secs = int(secs % 60)
    minutes = int(minutes % 60)
    hrs = int(hrs % 24)
    res = ""
    if days:
        res += "%id:" % days
    if hrs:
        res += "%ih:" % hrs
    if minutes:
        if hrs and minutes < 10:
            res += "0"
        res += "%im:" % minutes
    if minutes and secs < 10:
            res += "0"
    res += "%is" % secs
    return res

def parse_args():
    parser = argparse.ArgumentParser(prog='dwqc', description='dwq: disque-based work queue')

    parser.add_argument('-q', '--queue', type=str,
            help='queue name for jobs (default: \"default\")',
            default=os.environ.get("DWQ_QUEUE") or "default")

    parser.add_argument('-r', "--repo", help='git repository to work on', type=str,
            required="DWQ_REPO" not in os.environ, default=os.environ.get("DWQ_REPO"))

    parser.add_argument('-c', "--commit", help='git commit to work on', type=str,
            required="DWQ_COMMIT" not in os.environ, default=os.environ.get("DWQ_COMMIT"))

    parser.add_argument('-e', "--exclusive-jobdir", help='don\'t share jobdirs between jobs', action="store_true")

    parser.add_argument('-P', "--progress", help='enable progress output', action="store_true" )
    parser.add_argument('-R', "--report", help='report to disque queue', action="store", type=str)
    parser.add_argument('-v', "--verbose", help='enable status output', action="store_true" )
    parser.add_argument('-Q', "--quiet", help='don\'t print command output', action="store_true" )
    parser.add_argument('-m', "--maxfail", help='exit after more than <maxfail> jobs failed', type=int, default=sys.maxsize)
    parser.add_argument('-s', "--stdin", help='read from stdin', action="store_true" )
    parser.add_argument('-o', "--outfile", help='write job results to file', type=argparse.FileType('w'))
    parser.add_argument('-b', "--batch", help='send all jobs together', action="store_true")
    parser.add_argument('-S', "--subjob", help='pass job(s) to master instance, don\'t wait for completion', action="store_true")
    parser.add_argument('-E', "--env", help='export environment variable to client', type=str, action="append", default=[])
    parser.add_argument('-F', "--file", help='send file along with job', type=str, action="append", default=[])
    parser.add_argument('-a', "--asset", help='save specific asset', type=str, action="append", default=[])
    parser.add_argument('-A', "--asset-dir", help='save all assets', type=str, action="store")

    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    parser.add_argument('command', type=str, nargs='?')

    return parser.parse_args()

def get_env(env):
    result = {}
    for var in env:
        var = var.split("=", maxsplit=1)
        if len(var) == 1:
            val = os.environ.get(var[0])
            if val:
                var.append(val)
            else:
                continue
        result[var[0]] = var[1]

    return result

def create_body(args, command, options=None, parent_id=None):
    body = { "repo" : args.repo, "commit" : args.commit, "command" : command }
    if options:
        body["options"] = options

    if parent_id:
        body["parent"] = parent_id

    env = {}
    if args.env:
        env.update(get_env(args.env))

    env.update(options.get("env", {}))
    if env:
        body["env"] = env

    return body

def queue_job(jobs_set, queue, body, control_queues):
    timeout = body.get("options", {}).get("timeout")
    if timeout:
        job_id = Job.add(queue, body, control_queues, retry=timeout)
    else:
        job_id = Job.add(queue, body, control_queues)

    parent = body.get('parent')
    if parent:
        body = { 'parent' : parent, 'subjob' : job_id, 'unique' : os.environ.get("DWQ_JOB_UNIQUE") }
        Job.add(control_queues[0], body, None)

    else:
        jobs_set.add(job_id)

    return job_id

def vprint(*args, **kwargs):
    global verbose
    if verbose:
        print(*args, **kwargs)

verbose = False

def dict_addset(_dict, key, data):
    try:
        _dict[key].add(data)
    except KeyError:
        _dict[key] = {data}

def dict_dictadd(_dict, key):
    try:
        return _dict[key]
    except KeyError:
        _tmp = {}
        _dict[key] = _tmp
        return _tmp

def handle_assets(job, args):
    assets = job["result"].get("assets")
    asset_dir = args.asset_dir

    if assets:
        asset_map = {}
        for _asset in args.asset:
            _split = _asset.split(":")
            if len(_split) == 1:
                remote = _asset
                local = _asset
            elif len(_split) == 2:
                remote, local = _split
            else:
                print("dwqc: error: invalid asset spec \"%s\"" % _asset, file=sys.stderr)
                sys.exit(1)

            asset_map[remote] = local

        for remote, data in assets.items():
            local = asset_map.get(remote)
            if not local:
                if asset_dir:
                    local = os.path.realpath(os.path.join(asset_dir, remote))
                    if not local.startswith(asset_dir):
                        print("dwqc: warning: asset \"%s\" is not relative"
                              " to \"%s\", ignoring" % (remote, asset_dir),
                              file=sys.stderr)
                else:
                    print("dwqc: warning: ignoring asset \"%s\"" % remote, file=sys.stderr)
                    continue
            util.write_files({ local : data })

def main():
    global verbose
    args = parse_args()

    signal.signal(signal.SIGTERM, sigterm_handler)

    job_queue = args.queue

    Disque.connect(["localhost:7711"])

    if args.subjob:
        try:
            control_queue = os.environ["DWQ_CONTROL_QUEUE"]
        except KeyError:
            print("dwqc: error: --subjob specified, but DWQ_CONTROL_QUEUE unset.")
            sys.exit(1)

        try:
            parent_jobid = os.environ["DWQ_JOBID"]
        except KeyError:
            print("dwqc: error: --subjob specified, but DWQ_JOBID unset.")
            sys.exit(1)

    else:
        control_queue = "control::%s" % random.random()
        parent_jobid = None

    verbose = args.verbose

    if args.progress or args.report:
        start_time = time.time()

    if args.report:
        Job.add(args.report, { "status" : "collecting jobs" })

    try:
        file_data = util.gen_file_data(args.file)
    except util.GenFileDataException as e:
        print("dwqc: error processing --file argument:", e, file=sys.stderr)
        sys.exit(1)

    result_list = []
    try:
        jobs = set()
        batch = []
        if args.command and not args.stdin:
            options = {}
            if args.exclusive_jobdir:
                options.update({ "jobdir" : "exclusive" })
            if file_data:
                options["files"] = file_data
            queue_job(jobs, job_queue, create_body(args, args.command, options, parent_jobid), [control_queue])
        else:
            jobs_read = 0
            vprint("dwqc: reading jobs from stdin")
            for line in sys.stdin:
                line = line.rstrip()
                if args.stdin:
                    cmdargs = line.split(" ")
                    command = args.command
                    for i in range(0, len(cmdargs)):
                        command = command.replace("${%i}" % (i+1), cmdargs[i])
                else:
                    command = line

                tmp = command.split("###")
                command = tmp[0]
                options = {}
                if len(tmp) > 1:
                    command = command.rstrip()
                    try:
                        options = json.loads(tmp[1])
                    except json.decoder.JSONDecodeError:
                        vprint("dwqc: invalid option JSON. Skipping job.", file=sys.stderr)
                        continue

                _job_queue = options.get("queue", job_queue)

                if args.exclusive_jobdir:
                    options.update({ "jobdir" : "exclusive" })

                if file_data:
                    options["files"] = file_data

                if args.batch:
                    batch.append((_job_queue, create_body(args, command, options, parent_jobid), [control_queue]))
                else:
                    job_id = queue_job(jobs, _job_queue, create_body(args, command, options, parent_jobid), [control_queue])
                    vprint("dwqc: job %s command=\"%s\" sent to queue %s." % (job_id, command, _job_queue))
                    if args.progress:
                        print("")

                if args.progress or args.report:
                    jobs_read += 1
                    elapsed = time.time() - start_time

                    if args.progress:
                        print("\033[F\033[K[%s] %s jobs read" \
                            % (nicetime(elapsed), jobs_read), end="\r")

                    #if args.report:
                    #    Job.add(args.report, { "status" : "collecting jobs", "total" : jobs_read })

        _time = ""
        if args.batch and args.stdin:
            before = time.time()
            vprint("dwqc: sending jobs")
            for _tuple in batch:
                queue_job(jobs, *_tuple)
            _time = "(took %s)" % nicetime(time.time() - before)

            if args.report:
                Job.add(args.report, { "status" : "sending jobs"})

        if args.stdin:
            vprint("dwqc: all jobs sent.", _time)

        if args.subjob:
            if args.report:
                Job.add(args.report, { "status" : "done"})
            return

        if args.progress:
            vprint("")

        unexpected = {}
        early_subjobs = []
        total = len(jobs)
        done = 0
        failed = 0
        failed_expected = 0
        passed = 0
        subjobs = {}
        while jobs:
            _early_subjobs = early_subjobs or None
            early_subjobs = []

            for job in _early_subjobs or Job.wait(control_queue, count=128):
                #print(json.dumps(job, sort_keys=True, indent=4))
                subjob = job.get("subjob")
                if subjob:
                    parent = job.get("parent")
                    unique = job.get("unique")

                    _dict = dict_dictadd(subjobs, parent)
                    dict_addset(_dict, unique, subjob)

                else:
                    try:
                        job_id = job["job_id"]
                        jobs.remove(job_id)
                        done += 1
                        #if args.progress:
                        #    vprint("\033[F\033[K", end="")
                        #vprint("dwqc: job %s done. result=%s" % (job["job_id"], job["result"]["status"]))
                        if not args.quiet:
                            if args.progress:
                                print("\033[K", end="")
                            print(job["result"]["output"], end="")
                        _has_passed = job["result"]["status"] in { 0, "0", "pass" }
                        if _has_passed:
                            passed += 1
                            handle_assets(job, args)
                        else:
                            failed += 1
                            try:
                                if job["result"]["body"]["options"]["fail_ok"]:
                                    failed_expected += 1
                            except KeyError:
                                pass

                        if args.outfile:
                            result_list.append(job)

                        # collect subjobs started by this job instance, add to waitlist
                        unique = job["result"]["unique"]
                        _subjobs = subjobs.get(job_id, {}).get(unique, [])
                        for subjob_id in _subjobs:
                            try:
                                early_subjobs.append(unexpected.pop(subjob_id))
                            except KeyError:
                                pass
                            finally:
                                jobs.add(subjob_id)

                        total += len(_subjobs)

                        if args.progress or args.report:
                            elapsed = time.time() - start_time
                            per_job = elapsed / done
                            eta = (total - done) * per_job

                            if args.progress:
                                print("\r\033[K[%s] %s/%s jobs done (%s passed, %s failed.) " \
                                    "ETA:" % (nicetime(elapsed), done, total, passed, failed), nicetime(eta), end="\r")

                            if args.report:
                                Job.add(args.report, { "status" : "working", "elapsed" : elapsed, \
                                        "eta" : eta, "total" : total, "passed" : passed, "failed" : failed, "job" : job})

                        if not _has_passed:
                            if failed > failed_expected:
                                if (failed - failed_expected) > args.maxfail:
                                    print("dwqc: more than %i jobs failed. Exiting." % args.maxfail, file=sys.stderr)
                                    sys.exit(1)
                    except KeyError:
                        unexpected[job_id] = job

        if args.outfile:
            args.outfile.write(json.dumps(result_list))

        if args.progress:
            print("")

    except (KeyboardInterrupt, SystemExit):
        print("dwqc: cancelling...")
        Job.cancel_all(jobs)
        if args.report:
            Job.add(args.report, { "status" : "canceled"})
        sys.exit(1)

    Job.add(args.report, { "status" : "done"})

    if failed > failed_expected:
        sys.exit(1)



import json
import base64
import os

class GenFileDataException(Exception):
    pass

def gen_file_data(names, root=None):
    res = {}
    if not root:
        root = os.getcwd()
    for name in names:
        parts = name.split(":", maxsplit=1)
        if len(parts) == 2:
            local, remote = parts
            if remote.startswith("/"):
                raise GenFileDataException("file %s: remote name %s must be relative" % (local, remote))

            if ".." in remote:
                raise GenFileDataException("file %s: no \"..\" allowed in remote name %s" % (local, remote))
        else:
            local = name
            if name.startswith("/"):
                if not name.startswith(root + "/"):
                    raise Exception("file %s: remote path must be relative to cwd" % local)
                else:
                    remote = os.path.abspath(name)[len(root)+1:]
            else:
                remote = name

        if not local.startswith("/"):
            local = os.path.join(root, local)
        try:
            res[remote] = base64_file(local)
        except FileNotFoundError as e:
            raise GenFileDataException(e)

    return res

def base64_file(name):
    with open(name, "rb") as f:
        return base64.b64encode(f.read()).decode("ascii")

def write_files(data, workdir=None):
    data = data or {}
    for filename, filedata in data.items():
        if workdir:
            filename = os.path.join(workdir, filename)

        dirname = os.path.dirname(filename)
        if dirname:
            os.makedirs(dirname, exist_ok=True)

        with open(filename, "wb") as f:
            f.write(base64.b64decode(filedata))



import json
from pydisque.client import Client

disque = None

class Disque(object):
    def connect(servers):
        global disque
        disque = Client(servers)
        disque.connect()

    def get():
        global disque
        return disque

    def connected():
        global disque
        if not disque:
            return False
        if not disque.connected_node:
            return False
        return True

    def qscan(*args):
        queues = []
        n, _queues = disque.qscan(*args)
        for queue in _queues:
            queues.append(queue.decode('utf-8'))

        return queues

    def qstat(queues=None, *args):
        global disque

        if not queues:
            queues = Disque.qscan(*args)

        active = []
        queue_dict = {}

        for queue in queues:
            qstat_list = disque.qstat(queue) or []
            qstat = {}
            name = None
            while len(qstat_list):
                key = qstat_list[0].decode('ascii')
                val = qstat_list[1]
                qstat_list = qstat_list[2:]
                if isinstance(val, bytes):
                    val = val.decode('ascii')
                if key != 'name':
                    qstat[key] = val
                else:
                    name = val
            if name:
                queue_dict[name] = qstat

        return queue_dict

class Job(object):
    def __init__(s, job_id, body, queue_name, nacks, additional_deliveries):
        s.job_id = job_id
        s.body = body
        s.queue_name = queue_name
        s.nacks = nacks
        s.additional_deliveries = additional_deliveries
        s.control_queues = body.get("control_queues") or []

    def get(queues, timeout=None, count=None, nohang=False):
        global disque
        jobs = []
        _jobs = disque.get_job(queues, timeout=timeout, count=count, nohang=nohang, withcounters=True)
        for queue_name, job_id, json_body, nacks, additional_deliveries in _jobs:
            queue_name = queue_name.decode("ascii")
            job_id = job_id.decode("ascii")
            body = json.loads(json_body.decode("utf-8"))
            jobs.append(Job(job_id, body, queue_name, nacks, additional_deliveries))

        return jobs

    def working(s, status):
        global disque
        disque.working(s.job_id)

    def done(s, result):
        if s.control_queues:
            for queue in s.control_queues:
                if queue=="$jobid":
                    queue = s.job_id
                disque.add_job(queue, json.dumps({"job_id" : s.job_id, "state" : "done", "result" : result }))

        disque.ack_job(s.job_id)

    def add(queue, body, control_queues=None, **kwargs):
        if control_queues:
            body["control_queues"] = control_queues

        json_body = json.dumps(body)
        return disque.add_job(queue, json_body, **kwargs).decode("ascii")

    def nack(s):
        disque.nack_job(s.job_id)

    def cancel(s):
        disque.del_job(s.job_id)

    def wait(queue, count=None):
        _jobs = disque.get_job([queue], count=count)
        jobs = []
        for queue_name, job_id, json_body in _jobs:
            queue_name = queue_name.decode("ascii")
            job_id = job_id.decode("ascii")
            body = json.loads(json_body.decode("utf-8"))
            disque.fast_ack(job_id)
            jobs.append(body)
        return jobs

    def cancel_all(job_ids):
        disque.del_job(*tuple(job_ids))



#!/usr/bin/env python3

import argparse
import random
import sys

from dwq import Job, Disque
from dwq.version import __version__

def parse_args():
    parser = argparse.ArgumentParser(prog='dwqm', description='dwq: disque-based work queue (management tool)')

    subparsers = parser.add_subparsers(help='sub-command help')

    parser_queue = subparsers.add_parser('queue', help='queue help')
    parser_queue.set_defaults(func=queue)

    group = parser_queue.add_mutually_exclusive_group(required=True)
    group.add_argument('-l', '--list', help='list all disque queue(s)', action='store_true')
    group.add_argument('-d', '--drain', help='empty disque queue(s)', action='store_true')
    parser_queue.add_argument('queues', type=str, nargs='*')

    parser_control = subparsers.add_parser('control', help='control help')
    parser_control.set_defaults(func=control)
    group = parser_control.add_mutually_exclusive_group(required=True)
    group.add_argument('-l', '--list', help='list node(s)', action='store_true')
    group.add_argument('-p', '--pause', help='pause node(s)', action='store_true')
    group.add_argument('-r', '--resume', help='resume node(s)', action='store_true')
    group.add_argument('-s', '--shutdown', help='shutdown node(s)', action='store_true')
    group.add_argument('-P', '--ping', help='ping node(s)', action='store_true')

#    parser_control.add_argument('-f', '--force', help='don\'t wait for job completion on pause/shutdown',
#                                action='store_true')

    parser_control.add_argument('nodes', type=str, nargs='*')

    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    return parser.parse_args()

def print_queue(name, qstat):
    print("name:", name, "len:", qstat['len'], "blocked:", qstat['blocked'])

def listq(queues):
    Disque.connect(["localhost:7711"])

    qstat = Disque.qstat(queues)
    queues = sorted(qstat.keys())

    for name in queues:
        try:
            queue = qstat[name]
            print_queue(name, queue)
        except KeyError:
            print("invalid queue \"%s\"" % name)

def drain(queues):
    if not queues:
        print("dwqm: drain: no queues given.")
        sys.exit(1)

    Disque.connect(["localhost:7711"])
    disque = Disque.get()
    try:
        while True:
            jobs = Job.get(queues, count=1024, nohang=True)
            if not jobs:
                return

            job_ids = []
            for job in jobs:
                job_ids.append(job.job_id)

            disque.fast_ack(*job_ids)
    except KeyboardInterrupt:
        pass

def queue(args):
    if args.drain:
        drain(args.queues)
    elif args.list:
        listq(args.queues)

def control(args):
    Disque.connect(["localhost:7711"])

    jobargs = {}
#    if args.force:
#        jobargs["force"] = True

    if args.pause:
        control_cmd(args.nodes, "pause", **jobargs)
    elif args.resume:
        control_cmd(args.nodes, "resume")
    elif args.shutdown:
        control_cmd(args.nodes, "shutdown", **jobargs)
    elif args.ping:
        control_cmd(args.nodes, "ping")
    elif args.list:
        list_nodes()

def get_nodes():
    queues = Disque.qscan()
    nodes = []

    for queue in queues:
        if queue.startswith("control::worker::"):
            nodes.append(queue[17:])

    return nodes

def list_nodes():
    nodes = get_nodes()
    for node in sorted(nodes):
        print(node)

def control_cmd(nodes, cmd, **kwargs):
    control_queue = "control::%s" % str(random.random())
    job_ids = []
    for node in nodes:
        print("dwqm: sending \"%s\" command to node \"%s\"" % (cmd, node))
        job_id = control_send_cmd(node, cmd, control_queue, **kwargs)
        job_ids.append(job_id)

    while job_ids:
        for job in Job.wait(control_queue, count=len(job_ids)):
            job_id = job["job_id"]
            job_ids.remove(job_id)
            try:
                print("%s:" % job["result"]["worker"], job["result"].get("output"))
            except KeyError:
                pass

def control_send_cmd(worker_name, cmd, control_queue, **kwargs):
    body = { "control" : { "cmd" : cmd }}
    if kwargs:
        body["control"]["args"] = kwargs

    job_id = Job.add("control::worker::%s" % worker_name, body, [control_queue])
    return job_id

def main():
    args = parse_args()
    if "func" in args:
        args.func(args)
    else:
        print("dwqm: no command given")
        sys.exit(1)



#!/usr/bin/env python3
import json
import sys

from dwq import Job, Disque

def main():
    Disque.connect(["localhost:7711"])

    queues = sys.argv[1] or "default"
    try:
        while True:
            jobs = Job.wait(queues, count=16)
            for job in jobs:
                print(json.dumps(job, sort_keys=True, indent=4))
    except KeyboardInterrupt:
        pass

if __name__=="__main__":
    main()



#!/usr/bin/env python3

from setuptools import setup
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the version from file
with open(path.join(here, 'dwq/version.py'), encoding='utf-8') as f:
    _version = f.read()

version = _version.split("=")[1].lstrip().rstrip().lstrip('"').rstrip('"')

setup(
    name='dwq',
    version=version,

    description='dwq: a Disque based work queue',
    long_description=long_description,

    url='https://github.com/kaspar030/dwq',

    author='Kaspar Schleiser',
    author_email='kaspar@schleiser.de',

    license='GPLv3',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='distributed queue',
    packages=['dwq'],
    install_requires=['pydisque_dwq'],
    entry_points={
        'console_scripts': [
            'dwqc=dwq.dwqc:main',
            'dwqm=dwq.dwqm:main',
            'dwqw=dwq.dwqw:main',
        ],
    },
)



#!/usr/bin/env python3

import argparse
import os
import threading
import random
import time
import signal
import socket
import traceback
import multiprocessing
import shutil
import subprocess

import redis
from redis.exceptions import ConnectionError, RedisError

from dwq import Job, Disque

import dwq.cmdserver as cmdserver
from dwq.gitjobdir import GitJobDir

import dwq.util as util

from dwq.version import __version__

def sigterm_handler(signal, stack_frame):
    raise SystemExit()

def parse_args():
    parser = argparse.ArgumentParser(prog='dwqw', description='dwq: disque-based work queue (worker)')

    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    parser.add_argument('-q', '--queues', type=str,
            help='queues to wait for jobs (default: \"default\")', nargs='*', default=["default"])
    parser.add_argument('-j', "--jobs",
            help='number of workers to start', type=int, default=multiprocessing.cpu_count())
    parser.add_argument('-n', '--name', type=str,
            help='name of this worker (default: hostname)', default=socket.gethostname())
    parser.add_argument('-v', "--verbose", help='be more verbose', action="count", default=1)
    parser.add_argument('-Q', "--quiet", help='be less verbose', action="count", default=0)

    return parser.parse_args()

shutdown = False

active_event = threading.Event()

def worker(n, cmd_server_pool, gitjobdir, args, working_set):
    global active_event
    global shutdown
    print("worker %2i: started" % n)
    buildnum = 0
    while not shutdown:
        try:
            if not Disque.connected():
                time.sleep(1)
                continue
            while not shutdown:
                active_event.wait()
                jobs = Job.get(args.queues)
                for job in jobs:
                    if shutdown:
                        job.nack()
                        continue

                    if job.additional_deliveries > 2:
                        error = "too many deliveries (usual reason: timeout)"
                        vprint(2, "worker %2i: %s" % (n, error))
                        job.done({ "status" : "error", "output" : "dwqw: %s\n" % error,
                            "worker" : args.name, "runtime" : 0, "body" : job.body })
                        continue

                    buildnum += 1
                    working_set.add(job.job_id)
                    before = time.time()
                    vprint(2, "worker %2i: got job %s from queue %s" % (n, job.job_id, job.queue_name))

                    try:
                        repo = job.body["repo"]
                        commit = job.body["commit"]
                        command = job.body["command"]
                    except KeyError:
                        vprint(2, "worker %2i: invalid job json body" % n)
                        job.done({ "status" : "error", "output" : "worker.py: invalid job description" })
                        continue

                    vprint(2, "worker %2i: command=\"%s\"" % (n, command))

                    exclusive = None
                    try:
                        options = job.body.get("options") or {}
                        if options.get("jobdir") or "" == "exclusive":
                            exclusive = str(random.random())
                    except KeyError:
                        pass

                    unique = random.random()

                    _env = os.environ.copy()

                    try:
                        _env.update(job.body["env"])
                    except KeyError:
                        pass

                    _env.update({ "DWQ_REPO" : repo, "DWQ_COMMIT" : commit, "DWQ_QUEUE" : job.queue_name,
                                  "DWQ_WORKER" : args.name, "DWQ_WORKER_BUILDNUM" : str(buildnum),
                                  "DWQ_WORKER_THREAD" : str(n), "DWQ_JOBID" : job.job_id,
                                  "DWQ_JOB_UNIQUE" : str(unique), "DWQ_CONTROL_QUEUE" : job.body.get("control_queues")[0]})

                    workdir = None
                    workdir_error = None
                    try:
                        try:
                            workdir = gitjobdir.get(repo, commit, exclusive=exclusive or str(n))
                        except subprocess.CalledProcessError as e:
                            workdir_error = "dwqw: error getting jobdir. output: \n" + e.output.decode("utf-8")

                        if not workdir:
                            if job.nacks < options.get("max_retries", 2):
                                job.nack()
                                vprint(1, "worker %2i: error getting job dir, requeueing job" % n)
                            else:
                                job.done({ "status" : "error", "output" : workdir_error or "dwqw: error getting jobdir\n",
                                            "worker" : args.name, "runtime" : 0, "body" : job.body })
                                vprint(1, "worker %2i: cannot get job dir, erroring job" % n)
                            working_set.discard(job.job_id)
                            continue

                        util.write_files(options.get('files'), workdir)

                        # assets
                        asset_dir = os.path.join(workdir, "assets", "%s:%s" % (hash(job.job_id), str(unique)))
                        _env.update({ "DWQ_ASSETS" : asset_dir })

                        timeout = options.get("timeout", 300)

                        if timeout >= 8:
                            # send explit nack before disque times us out
                            # but only if original timeout is not too small
                            timeout -= 2

                        handle = cmd_server_pool.runcmd(command, cwd=workdir, shell=True, env=_env, start_new_session=True)
                        output, result = handle.wait(timeout=timeout)
                        if handle.timeout:
                            result = "timeout"
                            output = "dwqw: command timed out\n"

                        if (result not in { 0, "0", "pass" }) and job.nacks < options.get("max_retries", 2):
                            vprint(2, "worker %2i: command:" % n, command,
                                    "result:", result, "nacks:", job.nacks, "re-queueing.")
                            job.nack()
                        else:
                            runtime = time.time() - before

                            options = job.body.get("options")
                            if options:
                                options.pop("files", None)

                            # remove options from body if it is now empty
                            if not options:
                                job.body.pop("options", None)

                            _result = { "status" : result, "output" : output, "worker" : args.name,
                                        "runtime" : runtime, "body" : job.body, "unique" : str(unique) }

                            # pack assets
                            try:
                                asset_files = os.listdir(asset_dir)
                                if asset_files:
                                    _result.update({ "assets" :  util.gen_file_data(asset_files, asset_dir) })
                                    shutil.rmtree(asset_dir, ignore_errors=True)
                            except FileNotFoundError:
                                pass

                            job.done(_result)

                            vprint(2, "worker %2i: command:" % n, command,
                                    "result:", result, "runtime: %.1fs" % runtime)
                            working_set.discard(job.job_id)
                    except Exception as e:
                        if workdir:
                            gitjobdir.release(workdir)
                        raise e

                    gitjobdir.release(workdir)

        except Exception as e:
            vprint(1, "worker %2i: uncaught exception" % n)
            traceback.print_exc()
            time.sleep(10)
            vprint(1, "worker %2i: restarting worker" % n)

class SyncSet(object):
    def __init__(s):
        s.set = set()
        s.lock = threading.Lock()

    def add(s, obj):
        with s.lock:
            s.set.add(obj)

    def discard(s, obj):
        with s.lock:
            s.set.discard(obj)

    def empty(s):
        with s.lock:
            oldset = s.set
            s.set = set()
            return oldset

verbose = 0

def vprint(n, *args, **kwargs):
    global verbose
    if n <= verbose:
        print(*args, **kwargs)

def handle_control_job(args, job):
    global active_event
    global shutdown
    body = job.body
    status = 0
    result = ""

    try:
        control = body["control"]
        cmd = control["cmd"]
        if cmd == "shutdown":
            vprint(1, "dwqw: shutdown command received")
            result = "shutting down"
            shutdown = 1

        elif cmd == "pause":
            vprint(1, "dwqw: pause command received. pausing ...")
            active_event.clear()
            result = "paused"
        elif cmd == "resume":
            vprint(1, "dwqw: resume command received. resuming ...")
            active_event.set()
            result = "resumed"
        elif cmd == "ping":
            vprint(1, "dwqw: ping received")
            result="pong"
        else:
            vprint(1, "dwqw: unknown control command \"%s\" received" % cmd)

    except KeyError:
        vprint(1, "dwqw: error: invalid control job")

    control_reply(args, job, result, status)

    if shutdown:
        raise SystemExit()

def control_reply(args, job, reply, status=0):
    job.done({ "status" : status, "output" : reply, "worker" : args.name,
        "body" : job.body })

def main():
    global shutdown
    global verbose
    global active_event

    args = parse_args()
    verbose = args.verbose - args.quiet

    cmd_server_pool = cmdserver.CmdServerPool(args.jobs)

    signal.signal(signal.SIGTERM, sigterm_handler)

    _dir = "/tmp/dwq.%s" % str(random.random())
    gitjobdir = GitJobDir(_dir, args.jobs)

    servers = ["localhost:7711"]
    try:
        Disque.connect(servers)
        vprint(1, "dwqw: connected.")
    except:
        pass

    working_set = SyncSet()

    for n in range(1, args.jobs + 1):
        threading.Thread(target=worker, args=(n, cmd_server_pool, gitjobdir, args, working_set), daemon=True).start()

    active_event.set()

    try:
        while True:
            if not Disque.connected():
                try:
                    vprint(1, "dwqw: connecting...")
                    Disque.connect(servers)
                    vprint(1, "dwqw: connected.")
                except RedisError:
                    time.sleep(1)
                    continue

            try:
                control_jobs = Job.get(["control::worker::%s" % args.name])
                for job in control_jobs or []:
                    handle_control_job(args, job)
            except RedisError:
                pass

    except (KeyboardInterrupt, SystemExit):
        vprint(1, "dwqw: shutting down")
        shutdown = True
        cmd_server_pool.destroy()
        vprint(1, "dwqw: nack'ing jobs")
        jobs = working_set.empty()
        d = Disque.get()
        d.nack_job(*jobs)
        vprint(1, "dwqw: cleaning up job directories")
        gitjobdir.cleanup()

